; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         xdb.el
; RCS:          $Header: xdb.el,v 1.7 89/10/10 11:04:42 jbm Exp $
; Locker:       $Locker: jbm $
;
; Description:  GNU-Emacs interface to the HP xdb debugger.
;
; Starting up:
;
;     M-x xdb <ret>
;
;     You will be prompted for the path to an executable file, a
;     string of options to pass to xdb, and a list of name=value
;     pairs to be passed as the environment that xdb and the program
;     will run in.
;     
;     M-x describe-mode <ret> for a description of features.
;
;     Thank you for your support.
;
; Problems:
;
;     If you encounter an error which looks like an xdb-mode error, I
;     would appreciate your letting me know about it.
;
;     If the problem causes an emacs-lisp error, please try to
;     reproduce the error after setting both xdb-debug and
;     debug-on-error to t, then send me the contents of the
;     *Backtrace* and xdb-mode buffers.
;
; Disabilities:
;
;     Can't specify a core file which is not in the same directory
;     as the executable.
;
; (c) Copyright 1989, 1990, 1991, Hewlett-Packard Company, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'xdb)

;; GNU Emacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
;; to anyone for the consequences of using it or for whether it serves
;; any particular purpose or works at all, unless he says so in writing.
;; Refer to the GNU Emacs General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute GNU
;; Emacs, but only under the conditions described in the GNU Emacs
;; General Public License.  A copy of this license is supposed to have
;; been given to you along with GNU Emacs so you can know your rights and
;; responsibilities.  It should be in a file named COPYING.  Among other
;; things, the copyright notice and this notice must be preserved on all
;; copies.


;; User Variables


(defvar xdb-default-display-on t
  "* Determines whether xdb-mode will come up in display mode.
In display mode the source code is displayed in one window while the
interaction with xdb takes place in the xdb-mode buffer.  If this
variable is set nil, xdb-mode will work in line mode as if invoked
with the -L option.  See the xdb(1) man page.")

(defvar xdb-default-window-size 24
  "* Default height of the window displaying the xdb-mode buffer.
The height of the window will only be set when the buffer is initially
displayed, unless this is set to nil, in which case it will not be
sized at all.")

(defvar xdb-restore-windows t
  "* Flag indicating windows should be restored when xdb is quit.
The xdb-mode buffer is buried when xdb is quit.  It will still contain
all the debugger output until the next invocation of xdb-mode.")

(defvar xdb-core-backup nil
  "* If this flag is non-nil, xdb-mode will answer \"y\" when xdb asks
\"Do you want to save a backup copy of the core file?\", otherwise it
will answer \"n\"")

(defvar xdb-prompt-pattern "^[0-9]+> \\|^\\$ \\|^\\% "
  "* A regexp matching prompts of interactive programs run under xdb-mode.
The pattern should match a prompt at the beginning of the line.  This
is needed by xdb-mode to detect when the program is prompting the
user.  It defaults to a regexp matching some common prompts.  This
hack can probably be eliminated by a re-write of the process filters
and associated functions...")


;; Internal Variables


(defvar xdb-prompt-regexp
  "^>\\|^Really quit\\? \\|^Really terminate child process\\? \\|^File .+ exists, .+\\?\\|^Do you want to save a backup copy of the core file\\?\\|^Call procedure anyway\\?\\|^Invalid stack\\?"
  "A regexp for recognizing various xdb prompts.")


(defvar xdb-prompt-regexp-eol
  "^>$\\|^Really quit\\? $\\|^Really terminate child process\\? $\\|^File .+ exists, .+\\?$\\|^Do you want to save a backup copy of the core file\\? $\\|^Call procedure anyway\\?$\\|^Invalid stack\\?$"
  "Like xdb-prompt-regexp, but only matches at the end of a line.")


;;;
;;; This is a mucked up version of "^.+\.c: .+: [0-9]+.*: .+$"
;;; with the right \( \) pairs in it so I can use match-beginning
;;; and match-end.
;;;
(defvar xdb-file-line-regexp "^\\(.+\.c\\): \\(.+\\): \\([0-9]+\\).*: \\(.+\\)$"
  "A regexp used to detect when xdb has printed a line which contains the
file/proc/line information needed to display the file.  If this is changed,
function xdb-save-file-info needs to be changed as well.")


(defvar xdb-continued-command-regexp "^.:  .+\.c: .+: [0-9]+.*: .+$"
  "Similar in purpose to xdb-file-line-regexp, except this catches the
cases where just a newline is sent to xdb telling it to repeat the last
command.  In the single-step case, this prints the repeated command before
the actual line which is usually trapped by the aforementioned regexp.")


(defvar xdb-pid-regexp "^Starting process \\([0-9]+\\):"
  "A regexp used to figure out what the pid of xdb's child process is.")


(defvar xdb-child-terminated-regexp "^Child process terminated\\|^Terminating process"
  "A regexp used to figure out when xdb has lost its child process.")


(defvar xdb-current-buffer nil
  "The xdb-mode current buffer.")


(defvar xdb-command-name "xdb"
  "This is the command invoked to start the inferior process.")


(defvar xdb-required-options '("-L")
  "List of option strings that have to be used to run with xdb-mode.")


(defvar xdb-required-commands '("sm")
  "List of command strings that have to be passed to xdb initially to
make this whole deal work.")


(defvar xdb-debug nil
  "If set non-nil, all output from the xdb process (fed to xdb-filter) will
be printed to the xdb-mode buffer inside brackets in addition to the normal
output generated by the various filters.");


(defvar xdb-window-configuration nil
  "Window cofiguration to restore when I quit xdb-mode.")


(defvar xdb-mode-map nil
  "Keymap for xdb-mode.")


(defvar xdb-display-mode-map nil
  "Keymap for the xdb-mode display buffer where the source is displayed.")


(if xdb-mode-map
    nil
  (setq xdb-mode-map (make-keymap))
  ;(setq xdb-mode-map (copy-keymap shell-mode-map))
  (define-key xdb-mode-map "\C-l" 'xdb-refresh)
  (define-key xdb-mode-map "\C-m" 'xdb-send-input)
  (define-key xdb-mode-map "\C-a" 'xdb-beginning-of-line)
  (define-key xdb-mode-map "\C-cf" 'xdb-find-file)
  (define-key xdb-mode-map "\C-co" 'xdb-find-file-other-window)
  (define-key xdb-mode-map "\C-c1" 'xdb-find-file-one-window)
  (define-key xdb-mode-map "\C-cw" 'xdb-fix-window-size)
  (define-key xdb-mode-map "\C-cs" 'xdb-signal-program)
  (define-key xdb-mode-map "\C-ci" 'xdb-interrupt-program)
  (define-key xdb-mode-map "\C-cq" 'xdb-quit-program)
  (define-key xdb-mode-map "\C-ct" 'xdb-terminate-program)
  )


(if xdb-display-mode-map
    nil
  (setq xdb-display-mode-map (make-keymap))
  (suppress-keymap xdb-display-mode-map)
  )


(defun xdb-mode ()
  "
Major mode for interacting with an inferior xdb process.  This mode is
normally invoked by doing a M-x xdb.  Do \\[describe-function] on xdb
for a description of starting up xdb-mode.

The xdb-mode buffer is popped up after the xdb process is started.
The normal xdb startup sequence should appear a line at a time in the
buffer.  If the xdb-mode buffer does not appear in any windows (e.g.,
if you bury the buffer right after invoking xdb) an alert will be
printed to the echo area when the debugger is ready.

If xdb-default-display-on is non-nil, the source code will be
displayed in another window in xdb-display-mode.

The xdb process is run with the -L option.  As a result, any xdb
command may be entered in the xdb-mode buffer except the ones that
deal with terminal window features such as disassembly mode or
displaying the general registers.

Note that one of the limitations of this package is that prompt patterns
for interactive programs must be specified in the variable
xdb-prompt-pattern.

See the documentation for these user variables with \\[describe-variable]:

	xdb-default-display-on
	xdb-default-window-size
	xdb-restore-windows
	xdb-core-backup
	xdb-prompt-pattern

The following commands are available:

\\{xdb-mode-map}

\\[xdb-refresh] will refresh the display if it ever gets garbled.  It
recenters point in both the xdb-mode and display windows.  If you
switch away from the xdb-mode buffer and come back to it, this
function will redisplay the current source.

\\[yank] can be used as command recall to yank the previous command
onto the xdb-mode command line.  It uses the kill ring so \\[yank-pop]
can be used normally to further step backward through the kill ring
which holds the xdb-mode buffer's command history.

\\[xdb-find-file] will do a find-file on the current xdb-mode file.
Point is placed on the current line within that file for editing.  As
with xdb, xdb-mode may get confused if the edited file is saved to
disk and reference to that file is needed again.

\\[xdb-find-file-other-window] is like \\[xdb-find-file] except the
file is brought up in the other window, normally the display window.

\\[xdb-find-file-one-window] is like \\[xdb-find-file] except the file
will fill the entire screen (delete-other-windows is run).

\\[xdb-fix-window-size] will make the xdb-mode window the height found
in xdb-default-window-size, if it is set.  The the documentation on
that variable.

\\[xdb-signal-program] will prompt for the number of a signal and
then do a kill -<num> on the child process of xdb, the program which
is being debugged.  The signal number should be in the range of
numbers defined in signal(5).

\\[xdb-interrupt-program] is a special case of xdb-signal-program.  It
does the same thing as \\[xdb-signal-program] with an argument of 2
which is signal SIGINT.

\\[xdb-quit-program] is a special case of xdb-signal-program.  It
does the same thing as \\[xdb-signal-program] with an argument of 3
which is signal SIGQUIT.

\\[xdb-terminate-program] is a special case of xdb-signal-program.  It
does the same thing as \\[xdb-signal-program] with an argument of 15
which is signal SIGTERM.

"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'xdb-mode)
  (setq mode-name "Xdb")
  (setq mode-line-process '(": %s"))
  (use-local-map xdb-mode-map)
  (make-local-variable 'xdb-child-pid)             ; actual xdb child pid
  (make-local-variable 'xdb-current-context)	   ; xdb's line context
  (make-local-variable 'xdb-current-file)	   ; xdb's file
  (make-local-variable 'xdb-current-line)	   ; xdb's line
  (make-local-variable 'xdb-current-proc)	   ; xdb's procedure name
  (make-local-variable 'xdb-current-process)	   ; xdb process object
  (make-local-variable 'xdb-directory-list)	   ; -d directories
  (make-local-variable 'xdb-display-buffer)	   ; source buffer displayed
  (make-local-variable 'xdb-display-on)		   ; displaying source or not?
  (make-local-variable 'xdb-file-buffer-alist)	   ; hide buffers from user
  (make-local-variable 'xdb-last-input-end)	   ; for position tracking
  (make-local-variable 'xdb-last-input-start)	   ; for position tracking
  (make-local-variable 'xdb-line-accumulator)	   ; buffer for building lines
  (make-local-variable 'xdb-string-accumulator)	   ; for xdb-run-command output
  (make-local-variable 'xdb-user-filter)	   ; nil for xdb-run-command
  (make-local-variable 'xdb-window-size)	   ; resize on first display
  (make-local-variable 'xdb-window-sized)	   ; flag

  ;; Initialize.

  (setq xdb-child-pid nil)
  (setq xdb-current-context nil)
  (setq xdb-current-file "")
  (setq xdb-current-line -1)
  (setq xdb-current-proc "")
  (setq xdb-current-process nil)
  (setq xdb-directory-list nil)
  (setq xdb-display-buffer nil)
  (setq xdb-display-on xdb-default-display-on)
  (setq xdb-file-buffer-alist nil)
  (setq xdb-last-input-end (make-marker))
  (setq xdb-last-input-start (make-marker))
  (setq xdb-line-accumulator nil)
  (setq xdb-string-accumulator nil)
  (setq xdb-user-filter t)
  (setq xdb-window-size xdb-default-window-size)
  (setq xdb-window-sized nil)
  (run-hooks 'xdb-mode-hook))


(defun xdb-display-mode ()
  "
Major mode for interacting with xdb-mode.  This mode is not normally
invoked by the user, but is generated by xdb-mode to display source
code.

\\{xdb-display-mode-map}

"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'xdb-display-mode)
  (setq mode-name "Xdb Display")
  (make-local-variable 'xdb-process-buffer) ; this display's xdb-mode buffer
  (make-local-variable 'xdb-no-source)	    ; no source file found
  (setq xdb-process-buffer nil)
  (setq xdb-no-source nil)
  (setq mode-line-buffer-identification (list "Debug:%14b"))
  (setq buffer-read-only nil)
  (set-buffer-modified-p nil)
  (setq truncate-lines t)
  (use-local-map xdb-display-mode-map))


(defun xdb (path options environ)

  "Run xdb on PATH with OPTIONS in name=value pair environment ENVIRON.
Interactively, each of the arguments is a string which will be
prompted for in the minibuffer.  Xdb is run in buffer *xdb-PATH* in
which xdb-mode is invoked.  The directory where the program is located
becomes the initial working directory for xdb.  The name=value pairs
sets up the environment which will be added to the inherited
environment in which xdb (and therefore, the debugged program) will
run.  Once invoked, do M-x describe-mode for help on xdb-mode."
  (interactive "fRun xdb on file: \nsOptions: \nsEnvironment: ")
  (let ((arglist (break-string options "[ 	]+"))
	(envlist (break-string environ "[ 	]+"))
	(file (file-name-nondirectory path)))
    (setq xdb-window-configuration (current-window-configuration))
    (pop-to-buffer (concat "*xdb-" file "*"))
    (xdb-mode)
    (setq default-directory (file-name-directory (expand-file-name path)))
    (erase-buffer)
    (insert (format "Current directory is %s\n" default-directory))
    (insert (format "Starting xdb on %s -- %s\n" file (current-time-string)))
    (if (and (numberp xdb-window-size)
	     (not xdb-window-sized))
	(enlarge-window (- xdb-window-size (window-height))))
    (setq xdb-directory-list (xdb-make-directory-list arglist))
    (apply 'xdb-make-process (concat "xdb-" file) xdb-required-commands
	   (append envlist (list xdb-command-name) xdb-required-options arglist
		   (list file)))
    (setq xdb-current-process (get-buffer-process (current-buffer)))
    (set-process-filter xdb-current-process 'xdb-filter)
    (set-process-sentinel xdb-current-process 'xdb-sentinel))
    (setq xdb-current-buffer (current-buffer)))


(defun xdb-filter (proc string)
  "Main xdb-mode filter function.
Responsible for inserting output from xdb into the xdb-mode buffer,
and dealing with anything xdb prints out."
  (if (save-excursion
	(set-buffer (process-buffer proc))
	xdb-user-filter)
      (xdb-user-filter proc string)
    (xdb-internal-filter proc string)))


(defun xdb-user-filter (proc string)
  "Xdb-mode filter for stuff generated by xdb which the user invoked.
This includes all output in response to user commands typed in, as
well as commands bound to keys in the xdb-mode-keymap.  Also, this
catches the file/proc/line output."
  ;; This should be valid so long as they're only looking at one
  ;; xdb output at a time.
  (setq xdb-current-buffer (process-buffer proc))
  (let* ((prompt (string-match (concat xdb-prompt-regexp-eol
				       "\\|"
				       xdb-prompt-pattern)
			       string))
	 ;;
	 ;; The prompt is the only thing printed before an entire line
	 ;; of output is completed.  Since it's a special case, save
	 ;; it here and delete it from the string I'm actually
	 ;; scanning below.
	 ;;
	 (prompt-string (if prompt (substring string prompt)))
	 (fixed-string (if prompt
			   (substring string 0 prompt)
			 string)))
    (if xdb-debug
	(xdb-insert proc (format "[%s]" string)))
    (let ((newline nil)
	  (str fixed-string)
	  (tmp nil)
	  (startpos 0))
      (while (setq newline (string-match "\n" str startpos))
	(if xdb-debug
	    (xdb-insert proc "(1)"))
	(xdb-filter-line proc (concat xdb-line-accumulator
				      (substring str startpos newline)
				      "\n"))
	(setq xdb-line-accumulator nil)
	(setq startpos (if newline (1+ newline))))
      ;; Save anything left over for the next call to this filter.
      (if (> (length (setq tmp (substring str startpos)))
	     0)
	  (setq xdb-line-accumulator (concat xdb-line-accumulator tmp))))
    
    (if prompt
	(progn
	  (if (get-buffer-window xdb-current-buffer)
	      nil
	    (beep)
	    (message "Xdb: %s ready." (buffer-name xdb-current-buffer)))
	  (if xdb-debug
	      (xdb-insert proc "(2)"))
	  (xdb-insert proc prompt-string)))
    ))


(defun xdb-internal-filter (proc string)
  "Xdb-mode filter for output generated by xdb which xdb.el invoked.
This filter is used when the buffer local variable xdb-user-filter is
set to nil, which should only be for the duration of an xdb.el command.
Check to see if I want to clear xdb-string-accumulator before calling
this."
  (let* ((prompt (string-match (concat xdb-prompt-regexp-eol
				       "\\|"
				       xdb-prompt-pattern)
			       string))
	 (str (if prompt
		  (substring string 0 prompt)
		string)))
    (if xdb-debug (xdb-insert proc (format "{%s}" string)))
    ;;
    ;; Gather output until xdb is finished.
    ;;
    (while (not prompt)
      (setq xdb-string-accumulator
	    (concat xdb-string-accumulator str))
      (accept-process-output proc))
    ;;
    ;; Append what may have come before the prompt.
    ;;
    (setq xdb-string-accumulator
	  (concat xdb-string-accumulator str))))


(defun xdb-run-command (proc cmd)
  "Run an xdb command, transparent to the user.
Process PROC is sent string CMD with a newline appended.  Output is
first sent to the buffer local variable xdb-string-accumulator via
xdb-internal-filter, and is then returned.  This function waits for
output."
  (save-excursion
    (set-buffer (process-buffer proc))
    (setq xdb-user-filter nil)
    (setq xdb-string-accumulator nil)
    (process-send-string proc (concat cmd "\n"))
    (accept-process-output proc)
    (setq xdb-user-filter t)
    xdb-string-accumulator))


(defun xdb-filter-line (proc line)
  "Lower level xdb-mode filter function."

  ;; This one gets called when I have a whole line of output from xdb.
  ;; When in display mode (xdb-display-on non nil) a line matching the
  ;; pattern in xdb-file-line-regexp is not printed, but scanned to
  ;; save information about which file xdb is in and where the current
  ;; line in that file is.  This information is used to display the
  ;; file in another window.

  ;; Note that the save-excursion to the xdb-current-buffer limits me
  ;; to starting up only one xdb process at a time since it's a global
  ;; variable keeping track of the current one.

;;;; I might want to figure out some way of blocking the startup of
;;;; another xdb process until one has completely started.  Maybe
;;;; something like checking if the initial file has been set yet.

  ;; Save the process id.
  (if (string-match xdb-pid-regexp line)
      (save-excursion
	(set-buffer xdb-current-buffer)
	(setq xdb-child-pid
	      (substring line (match-beginning 1) (match-end 1)))))

  ;; Check for dead child.
  (if (string-match xdb-child-terminated-regexp line)
      (save-excursion
	(set-buffer xdb-current-buffer)
	(setq xdb-child-pid nil)))

  (if (save-excursion
	(set-buffer xdb-current-buffer)
	xdb-display-on)
      ;;
      ;; Test for the continued-command output lines.  Note that this
      ;; has to come before the xdb-file-line-regexp comparison since
      ;; it'll trap anything this one does.
      ;;
      (if (string-match xdb-continued-command-regexp line)
	  (xdb-save-file-info (substring line
					 (progn
					   (string-match "^.:  " line nil)
					   (match-end 0))))
	;;
	;; Didn't get a continued-command line, try the regular ones.
	;;
	(if (string-match xdb-file-line-regexp line)
	    (xdb-save-file-info line)
	  ;;
	  ;; Didn't detect anything xdb-mode is worried about, just
	  ;; print the line.
	  ;;
	  (xdb-insert proc line)))
    ;;
    ;; I'm not in the xdb-mode buffer.  Just print the line.
    ;;
    (xdb-insert proc line)))
	

(defun xdb-save-file-info (line)
    "Stashes information about the current file and line.  Redisplays the
file if xdb-display-on is non-nil.  Note that xdb-display-on is going
to be checked for the current buffer.  So if the user switches away from
that buffer, the information still gets saved, but the display won't happen."
  (let (newfile newproc newline newcontext index)
    (string-match xdb-file-line-regexp line 0) ; set up for match-*
    (save-excursion
      (set-buffer xdb-current-buffer)

      ;; File
      (setq newfile (substring line (match-beginning 1) (match-end 1)))
      (if (string-equal newfile xdb-current-file)
	  (setq newfile "")
	(setq xdb-current-file newfile))

      ;; Procedure
      (setq newproc (substring line (match-beginning 2) (match-end 2)))
      (if (string-equal newproc xdb-current-proc)
	  (setq newproc nil)
	(setq xdb-current-proc newproc))

      ;; Line
      (setq newline (string-to-int (substring line (match-beginning 3)
					      (match-end 3))))
      (if (= newline xdb-current-line)
	  (setq newline nil)
	(setq xdb-current-line newline))

      ;; Context
      (setq newcontext (substring line (match-beginning 4) (match-end 4)))
      (if (string-equal newcontext xdb-current-context)
	  (setq newcontext nil)
	(setq xdb-current-context newcontext)))

    ;; Now update
    (if (and (eq major-mode 'xdb-mode)
	     xdb-display-on)
	;; Check for source.
	(if (string-match "No Source"
			  (if newcontext newcontext xdb-current-context))
	    (let* ((buf (get-buffer-create (concat " " newfile)))
		   (window (display-buffer buf t)))
	      (setq xdb-display-buffer buf)
	      (save-excursion
		(set-buffer buf)
		(xdb-display-mode)
		;; Set the source here so it can be displayed if the
		;; D command is given.
		(setq xdb-no-source newfile)
		(setq xdb-process-buffer xdb-current-buffer)
		(erase-buffer)
		;; Stick a message in there
		(save-window-excursion
		  (select-window (get-buffer-window buf))
		  (insert-char ?\n (/ (window-height) 2))
		  (insert-char ?  (- (/ (window-width) 2) 10)))
		(insert "No Source")
		(goto-char (point-min))
		(setq buffer-read-only t)))
	  ;; Ah, got it.
	  (cond ((not (string-equal newfile ""))
		 (xdb-display-file newfile newline))
		(newline (xdb-display-line newline))))
    )))


(defun xdb-insert (proc string)
  "Insert stuff in the xdb-mode buffer.  Protect from errors in case
this is being done in the background while the user does other things
as xdb is loading."
  (let ((old-buffer (current-buffer)))
    (set-buffer (process-buffer proc))
    (unwind-protect
	(let ((pmark (process-mark proc)))
	  (save-excursion
	    (goto-char pmark)
	    (insert string)
	    (set-marker pmark (point)))
	  (goto-char pmark))
      (set-buffer old-buffer))))


(defun xdb-sentinel (proc msg)
  "This function is executed when the xdb process is quit, signalled,
or when the xdb-mode buffer has been killed."
  (cond ((null (buffer-name (process-buffer proc)))
	 ;; buffer killed
	 ;; Stop displaying an arrow in a source file.
	 (setq overlay-arrow-position nil)
	 (set-process-buffer proc nil)
	 (xdb-clear-lookup-list xdb-file-buffer-alist))
	((memq (process-status proc) '(signal exit))
	 ;; Stop displaying an arrow in a source file.
	 (setq overlay-arrow-position nil)
	 ;; Fix the mode line.
	 (setq mode-line-process
	       (concat ": "
		       (symbol-name (process-status proc))))
	 (xdb-clear-lookup-list xdb-file-buffer-alist)
	 (xdb-cleanup-buffer)
	 (if xdb-restore-windows
	     (set-window-configuration xdb-window-configuration))
	 )))


(defun xdb-cleanup-buffer ()
  "Writes final message in xdb-mode buffer and gets rid of its process."
  (let* ((obuf (current-buffer)))
    ;; save-excursion isn't the right thing if
    ;;  process-buffer is current-buffer
    (unwind-protect
	(progn
	  ;; Write a message indicating we're done.
	  (set-buffer (process-buffer proc))
	  (set-buffer-modified-p (buffer-modified-p))
	  (if (not (eobp))
	      (goto-char (point-max)))
	  (insert ?\n mode-name " " msg)
	  ;; If buffer and mode line will show that the process
	  ;; is dead, we can delete it now.  Otherwise it
	  ;; will stay around until M-x list-processes.
	  (delete-process proc)
	  (if xdb-restore-windows
	      (bury-buffer (current-buffer))))
      ;; Restore old buffer, but don't restore old point
      ;; if obuf is the xdb buffer.
      (set-buffer obuf))))


(defun xdb-display-file (file line)
  "This gets called when xdb has moved to a new file, and I have to
display the new position.  Should not get called when the file has not
chaged and only the line number has."
  (let* ((buffer (xdb-lookup-file file))
	 (window (display-buffer buffer t))
	 (pos nil))
    (select-window (get-buffer-window xdb-current-buffer))
    (setq xdb-display-buffer buffer)
    (xdb-set-overlay buffer line)
    (set-window-point window overlay-arrow-position)))

    
(defun xdb-display-line (line)
  "This gets called when xdb has moved to a different line in the same file
it was displaying previously.  Advance the overlay arrow and maybe recenter
the window."
  (let* ((buffer xdb-display-buffer)
	 (window (or (get-buffer-window buffer)
		     (pop-to-buffer buffer))))
    (select-window (get-buffer-window xdb-current-buffer))
    (xdb-set-overlay buffer line)
    (set-window-point window overlay-arrow-position)))


(defun xdb-set-overlay (buffer line)
  "Give a BUFFER and LINE number, set the overlay arrow."
  (save-excursion
    (set-buffer buffer)
    (goto-line line)
    (setq pos (point))
    (setq overlay-arrow-string "->")
    (or overlay-arrow-position
	(setq overlay-arrow-position (make-marker)))
    (set-marker overlay-arrow-position (point) (current-buffer))))


(defun xdb-refresh ()
  "In case the file display gets garbled, this re-displays the current
file and sets the overlay arrow to the current line.  Also recenters
the xdb-mode buffer."

;;;
;;; This should take care of the case where a line with file info was
;;; read, but no display was made because the current buffer wasn't the
;;; xdb-mode one.  That way I can hide it while it's starting up and
;;; when I switch back to it after it's finished all I have to do is hit
;;; C-l (or whatever) and the display comes up.
;;;

  (interactive)
  (let ((file xdb-current-file)
	(line xdb-current-line)
	(owin (get-buffer-window (current-buffer))))
    (xdb-display-file file line)
    (select-window (get-buffer-window xdb-display-buffer))
    (recenter)
    (select-window owin)
    (recenter)))


(defun xdb-make-directory-list (arglist)
  "Returns the directories given with the -d options of the argument
list, for setting in xdb-directory-list."
  (cond ((null arglist) nil)
	(t
	 (if (string-equal (car arglist) "-d")
	     (cons (car (cdr arglist))
		   (xdb-make-directory-list (cdr (cdr arglist))))
	   ))))

      
(defun xdb-lookup-file (file)
  "Given FILE name, return a buffer containing that file.
Looks in xdb-file-buffer-alist first, and if it's not there, looks in
the xdb-mode buffer's current directory for it.  If it's not found
there, do an xdb \"ld\" command and look in each of the directories
listed there.  If that doesn't work, return nil."
  (let ((lookup (assoc file (save-excursion
			      (set-buffer xdb-current-buffer)
			      xdb-file-buffer-alist)))
	(path nil))
    (if lookup
	(car (cdr lookup))
      ;;
      ;; Didn't find it, check the current directory.
      ;;
      (setq path (concat default-directory file))
      (if (file-readable-p path)
	  (xdb-make-lookup-list-entry path)
	;;
	;; File not in current directory, do search.
	;;
	(let ((found nil)
	      (dirlist nil)
	      (dir nil))
	  (save-excursion
	    (set-buffer xdb-current-buffer)
	    (setq dirlist xdb-directory-list)
	    (while (and dirlist (not found))
	      (setq dir (directory-no-slash (car dirlist)))
	      (if (file-readable-p (setq dir (concat dir "/" file)))
		  (setq found dir))))
	  (if (not found)
		nil
	    (xdb-make-lookup-list-entry dir))
	  )))))


(defun xdb-make-lookup-list-entry (path)
  "Given a PATH to a file, make an entry for the file in
xdb-file-buffer-alist.  Returns the buffer created."

  ;; This buffer local variable is an alist holding buffer entries for
  ;; files that have been displayed by xdb-mode.  This is done to keep
  ;; the buffers invisible to the user.  Each (FILE BUFFER . PATH)
  ;; element is accessible with assoc by the FILE key which is the
  ;; name of the .c file that is held by BUFFER.  This function
  ;; creates a buffer with a name generated by prefixing a space to
  ;; the basename of path.

  (let* ((name (concat " " (file-name-nondirectory path)))
	 (buffer (get-buffer-create name)))
    (save-excursion
      (set-buffer buffer)
      (xdb-display-mode)
      (setq xdb-process-buffer xdb-current-buffer)
      (erase-buffer)
      (insert-file-contents path)
      (xdb-number-lines)
      (setq buffer-read-only t))
    (save-excursion
      (set-buffer xdb-current-buffer)
      (setq xdb-display-buffer buffer)
      (setq xdb-file-buffer-alist
	    (cons (cons (file-name-nondirectory path)
			(cons buffer path))
		  xdb-file-buffer-alist)))
    buffer))


(defmacro xdb-clear-lookup-list (alist)
  "Destroys buffers and clears an ALIST of the form of xdb-file-buffer-alist."

  ;; This macro is generalized for any association list of the same
  ;; form as the buffer-local variable xdb-file-buffer-alist.  I'll
  ;; leave it to whatever function is calling it to make sure it's in
  ;; the right buffer context before trying to destroy its buffers.

  (` (progn (mapcar (function
		     (lambda (el)
		       (let ((buffer (car (cdr el))))
			 (kill-buffer buffer))))
		    (, alist))
	    (setq (, alist) nil))))


(defun break-string (string pattern)
  "Break up STRING into a list of strings at regexp PATTERN delimiter."
  (let (match)
    (cond ((string-equal string "") nil)
	  (string (cons
		   (substring string 0 
			      (if (setq match (string-match pattern string))
				  (match-beginning 0)))
		   (break-string (if match
				     (substring string (match-end 0)))
				 pattern)))
	  (t nil))))


(defun xdb-make-process (name cmds &rest args)
  "Create inferior xdb process named NAME.  Initial commands given on
stdin are in list CMDS.  The rest of the arguments are the tokens in
the command to be passed to env(1).  This function also figures out if
it has to answer the \"save backup?\" question."
  (let ((buffer (current-buffer))
	proc status size)
    (setq proc (get-buffer-process buffer))
    (if proc (setq status (process-status proc)))
    (if (and (file-exists-p "core")
	     (file-newer-than-file-p "core" name))
	  (setq cmds (cons (if xdb-core-backup "y" "n") cmds)))
    (save-excursion
      (set-buffer buffer)
      (if (memq status '(run stop))
	  nil
	(if proc
	    (delete-process proc))
	(setq proc (apply 'start-process name buffer "/bin/env" args))
	(if cmds
	    (mapcar (function
		     (lambda (cmd)
		       (process-send-string proc
					    (concat cmd "\n"))))
		    cmds))
	(setq name (process-name proc)))
      (goto-char (point-max))
      (set-marker (process-mark proc) (point)))
    buffer))


(defun xdb-send-input ()
  "Send input to xdb inferior process.
Input is only accepted when point is at the end of the buffer and
there is an xdb prompt at the beginning of the line.  Input is saved
to the kill ring for yanking later."
  (interactive)
  (or (get-buffer-process (current-buffer))
      (error "Current buffer has no process"))
  ;; Do nothing if there's no prompt yet.
  (if (and (bolp) (eobp))
      (message "Xdb: not ready for input, wait for prompt.")
    (end-of-line)
    (if (not (eobp))
	(message "Xdb: input only accepted and end of buffer.")
      (move-marker xdb-last-input-start
		   (process-mark (get-buffer-process (current-buffer))))
      (insert ?\n)
      (move-marker xdb-last-input-end (point))
      (let ((process (get-buffer-process (current-buffer))))
	;; Stick into the kill ring to yank later.  Leave off the newline.
	(copy-region-as-kill xdb-last-input-start (1- xdb-last-input-end))
	;; Check for directory commands.
	(xdb-check-directory-command (buffer-substring xdb-last-input-start
						       (1- xdb-last-input-end)))
	(process-send-region process xdb-last-input-start xdb-last-input-end)
	(set-marker (process-mark process) (point)))
      )))


(defun xdb-check-directory-command (str)
  "Check STR for an xdb directory command and add to the xdb-mode list
if needed.  If the display buffer doesn't have a file associated with
it, the new directory is checked for the source and xdb-display-file
done if needed."
  (if (string-match "[ 	]*D[ 	]+\"\\(.+\\)\"[ 	]*" str)
      (let ((dir (substring str (match-beginning 1) (match-end 1)))
	    (disp xdb-display-buffer))
	(setq xdb-directory-list (cons dir xdb-directory-list))
	;;
	;; xdb-no-source in the display buffer might have the file name
	;; if it isn't displaying source code.
	;;
	(let ((file (save-excursion (set-buffer xdb-display-buffer)
				    xdb-no-source)))
	  (xdb-display-file file xdb-current-line)))))

				       
(defun directory-no-slash (dir)
  "If DIR has a trailing slash, return it without the slash."
  (let ((dirlen (length dir)))
    (if (equal ?/ (aref dir (1- dirlen)))
	(substring dir 0 (- dirlen 2))
      dir)))

  
(defun xdb-number-lines ()
  "Adds line numbers to the current buffer.  Leaves enough space at the
front of the line for the overlay arrow."
  (let ((point (point-marker))
	(line-number 1)
	(count 0)
	format-string
	nlines)
    (untabify (point-min) (point-max))
    (setq nlines (count-lines (point-min) (point-max)))
    (setq format-string (cond ((< nlines 100) "   %2d ")
			      ((< nlines 1000) "   %3d ")
			      ((< nlines 10000) "   %4d ")
			      (t "   %7d ")))
    (goto-char (point-min))
    (while (<= count nlines)
      (insert-before-markers (format format-string line-number))
      (setq insert-end (point))
      (setq line-number (1+ line-number)
	    count (1+ count))
      (forward-line 1))))


(defun xdb-beginning-of-line ()
  "Like beginning-of-line, but move past anything in xdb-prompt-regexp."
  (interactive)
  (beginning-of-line)
  (re-search-forward (concat xdb-prompt-regexp "\\|" xdb-prompt-pattern)
		     (save-excursion (end-of-line) (point))
		     t))


(defun xdb-find-file ()
  "Do a find-file on the current xdb-mode file."
  (interactive)
  (switch-to-buffer (xdb-find-file-noselect)))


(defun xdb-find-file-other-window ()
  "Do a find-file on the current xdb-mode file in the other window."
  (interactive)
  (switch-to-buffer-other-window (xdb-find-file-noselect)))


(defun xdb-find-file-one-window ()
  "Do a find-file on the current xdb-mode file making it the only window."
  (interactive)
  (switch-to-buffer (xdb-find-file-noselect))
  (delete-other-windows))


(defun xdb-find-file-noselect ()
  "Read the current xdb-mode file into a buffer and return it.
Calls find-file-noselect after looking up the file path in
xdb-file-buffer-alist."
  (let ((items (xdb-lookup-file-info))
	path line buf)
      (if (not items)
	  (error "Xdb: path no known for %s" xdb-current-file))
      (setq path (nth 0 items))
      (setq line (nth 1 items))
      (save-excursion
	(set-buffer (setq buf (find-file-noselect path)))
	(goto-line line))
      buf))


(defun xdb-get-buffer ()
  "Returns the current xdb-mode buffer or the one associated with
an xdb-display-mode buffer, or calls error."
  (cond ((eq major-mode 'xdb-mode) (current-buffer))
	((eq major-mode 'xdb-display-mode) xdb-process-buffer)
	(t (error "Not an xdb-mode buffer."))))


(defun xdb-lookup-file-info (&optional file)
  "Lookup facility for the xdb-mode file buffer alist.
In an xdb-mode or xdb-display-mode buffer, returns a list of the path
to the current file, and the current line number in that file, and the
buffer object containing the file.  Optional argument FILE will lookup
the file instead of the current xdb-mode one, and only the path and
the buffer will be returned.  If the file isn't found in
xdb-file-buffer-alist, nil is returned."
  (let (lookup)
    (save-excursion
      (set-buffer (xdb-get-buffer))
      (if file
	  (if (setq lookup (assoc file xdb-file-buffer-alist))
	      (list (cdr (cdr lookup)) (car (cdr lookup)))
	    nil)
	(if (setq lookup (assoc xdb-current-file xdb-file-buffer-alist))
	  (list (cdr (cdr lookup)) xdb-current-line (car (cdr lookup)))
	nil)))))


(defun xdb-fix-window-size ()
  "Make the current xdb-mode window the height in xdb-window-size."
  (interactive)
  (save-excursion
    (set-buffer (xdb-get-buffer))
    (if (numberp xdb-window-size)
	(progn
	  (delete-other-windows)
	  (split-window-vertically xdb-window-size)
	  (xdb-refresh)))))


(defun xdb-signal-program (sig)
  "Send SIG (an integer) signal to the child process of xdb.
Interactively, this prompts for a signal number.  The range of the
value is checked for valid HP-UX signal number.  The child process is
determined by scanning for \"Starting process\" in the xdb output, and
saving the pid."
  (interactive "nSignal Number: ")
  (let* ((buf (xdb-get-buffer))
	 (proc (save-excursion
		 (set-buffer buf)
		 (if (not xdb-child-pid)
		     (error "Xdb: no process being debugged."))
		 xdb-child-pid)))
    (if (not (and (> sig 0) (< sig 31)))
	(error "Xdb: signal number out of range: %d" sig))
    (setq obj (start-process "xdb-signal" nil "kill" (int-to-string (- sig))
			     xdb-child-pid))
    (set-process-sentinel obj 'xdb-signal-sentinel)))


(defun xdb-signal-sentinel (proc msg)
  "Process sentinel for xdb-signal-program."
  (cond ((eq (process-status proc) 'exit)
	 (setq xdb-child-pid nil)
	 (if (equal 0 (process-exit-status proc))
	     (message "Xdb: signal sent.")
	   (beep t)
	   (message "Xdb: non-zero return on kill: %s" msg)))
	((eq (process-status proc) 'signal)
	 (message "Xdb: kill command signalled (%d): %s"
		  (process-exit-status proc) msg))
	(t nil)))


(defun xdb-interrupt-program ()
  "Send kill -2 to the program being debugged."
  (interactive)
  (xdb-signal-program 2))

(defun xdb-quit-program ()
  "Send kill -3 to the program being debugged."
  (interactive)
  (xdb-signal-program 3))

(defun xdb-terminate-program ()
  "Send kill -15 to the program being debugged."
  (interactive)
  (xdb-signal-program 15))
